{{
***************************
* Disassembler for        *
* Propeller assembly code *
*  (C) 2007 Thomas Kliche *
***************************

Revision History
----------------
 1.1 on 29 March 2007
     - fixed in getNextPC when debug step into at jmpret
     - fixed in getConditionStr, ifNC was missing
     - removed object dependency
     - added method getPrintableStr to format data as printable characters
       which is useful in hexadecimal view
     - some code optimizations

 1.0 on 18 March 2007
     - first public version

...................................................................................................

This is a disassembler to decode assembly instructions of the Propeller. You can use it like
following example.

OBJ
  screen : "TV_Text"
  disasm : "Disassembler"

PUB start
  screen.start(12)

PUB showInstruction(addr, instr)
'addr  - the address of instruction
'instr - the long of assembly instruction

  disasm.decode(instr)                 'this must be always called first
  screen.hex(addr, 3)
  screen.out(" ")
  screen.str(disasm.getConditionStr)   'show condition if available
  screen.out(" ")
  screen.str(disasm.getInstructionStr) 'show instruction mnemonic
  screen.out(" ")
  screen.str(disasm.getDestSrcStr)     'show destination and source of instruction
  screen.out(" ")
  screen.str(disasm.getEffectStr)      'show effects if available


The disassembler depends on the object AsmDebug.spin. Please keep in mind to update the OBJ section
of this file to your named debug file. Otherwise it will not work.

---------------------------------------------------------------------------------------------------
}}
CON

  'command for communication with debug kernel
  CMD_RDCOGDATA   = 3                                   'read data from Cog memory

  FLAG_I         = %0001
  FLAG_R         = %0010
  FLAG_C         = %0100
  FLAG_Z         = %1000

  MSK_ADDR       = $1FF
  MSK_COND_FLAGS = $F
  MSK_OPCODE     = $3F

  'some operation codes
  OPC_CALL       = $17
  OPC_AND        = $18
  OPC_SUB        = $21
  OPC_CMPS       = $30
  OPC_CMPSX      = $31
  OPC_SUBX       = $33

  TYPE_BYTE      = 0
  TYPE_WORD      = 1
  TYPE_LONG      = 3


VAR

  byte  DestAddrVisible
  byte  SrcAddrVisible
  byte  EffectVisible

  byte  Effect[9]
  byte  DestAddr[5]
  byte  SrcAddr[6]
  byte  Temp[10]

  byte  InstrOpcode
  byte  InstrCond
  byte  InstrFlags
  word  InstrDest
  word  InstrSrc


PUB decode(instrCode)

  InstrSrc := instrCode & MSK_ADDR
  InstrDest := (instrCode >>= 9) & MSK_ADDR
  InstrCond := (instrCode >>= 9) & MSK_COND_FLAGS
  InstrFlags := (instrCode >>= 4) & MSK_COND_FLAGS
  InstrOpcode := (instrCode >> 4) & MSK_OPCODE


PUB getConditionStr : condStr

  condStr := lookupz(InstrCond: @never_always, @ifNZaNC, @ifZaNC, @ifNC, {
}                               @ifNZaC, @ifNZ, @ifZneC, @ifNZoNC, {
}                               @ifZaC, @ifZeqC, @ifZ, @ifZoNC,    {
}                               @ifC, @ifNZoC, @ifZoC, @never_always)


PUB getInstructionStr : instruction

  DestAddrVisible := TRUE
  SrcAddrVisible := TRUE
  EffectVisible := TRUE
  case InstrOpcode
    $00:
      EffectVisible := FALSE
      if InstrCond
        if (InstrFlags & FLAG_R) == FLAG_R
          bytemove(@Temp, string("rdbyte"), 7)
        else
          bytemove(@Temp, string("wrbyte"), 7)
      else
        bytemove(@Temp, string("nop"), 4)
        DestAddrVisible := FALSE
        SrcAddrVisible := FALSE

    $01:
      EffectVisible := FALSE
      if (InstrFlags & FLAG_R) == FLAG_R
        bytemove(@Temp, string("rdword"), 7)
      else
        bytemove(@Temp, string("wrword"), 7)

    $02:
      EffectVisible := FALSE
      if (InstrFlags & FLAG_R) == FLAG_R
        bytemove(@Temp, string("rdlong"), 7)
      else
        bytemove(@Temp, string("wrlong"), 7)

    $03:
      EffectVisible := FALSE
      if (InstrFlags & FLAG_I) == FLAG_I
        SrcAddrVisible := FALSE
        case InstrSrc & 7
          0:
            bytemove(@Temp, string("clkset"), 7)

          1:
            bytemove(@Temp, string("cogid"), 6)

          2:
            bytemove(@Temp, string("coginit"), 8)
            EffectVisible := TRUE

          3:
            bytemove(@Temp, string("cogstop"), 8)

          4:
            bytemove(@Temp, string("locknew"), 8)
            EffectVisible := TRUE

          5:
            bytemove(@Temp, string("lockret"), 8)

          6:
            bytemove(@Temp, string("lockset"), 8)
            EffectVisible := TRUE

          7:
            bytemove(@Temp, string("lockclr"), 8)
            EffectVisible := TRUE
      else
        bytemove(@Temp, string("hubop"), 6)

    $04:
      bytemove(@Temp, string("mul"), 4)

    $05:
      bytemove(@Temp, string("muls"), 5)

    $06:
      bytemove(@Temp, string("enc"), 4)

    $07:
      bytemove(@Temp, string("ones"), 5)

    $08:
      bytemove(@Temp, string("ror"), 4)

    $09:
      bytemove(@Temp, string("rol"), 4)

    $0A:
      bytemove(@Temp, string("shr"), 4)

    $0B:
      bytemove(@Temp, string("shl"), 4)

    $0C:
      bytemove(@Temp, string("rcr"), 4)

    $0D:
      bytemove(@Temp, string("rcl"), 4)

    $0E:
      bytemove(@Temp, string("sar"), 4)

    $0F:
      bytemove(@Temp, string("rev"), 4)

    $10:
      bytemove(@Temp, string("mins"), 5)

    $11:
      bytemove(@Temp, string("maxs"), 5)

    $12:
      bytemove(@Temp, string("min"), 4)

    $13:
      bytemove(@Temp, string("max"), 4)

    $14:
      bytemove(@Temp, string("movs"), 5)

    $15:
      bytemove(@Temp, string("movd"), 5)

    $16:
      bytemove(@Temp, string("movi"), 5)

    $17:
      EffectVisible := FALSE
      if (InstrFlags & FLAG_I) == FLAG_I
        DestAddrVisible := FALSE
        if (InstrFlags & FLAG_R) == FLAG_R
          bytemove(@Temp, string("call"), 5)
        elseif InstrSrc == 0
          bytemove(@Temp, string("ret"), 4)
        else
          bytemove(@Temp, string("jmp"), 4)
      else
        if (InstrFlags & FLAG_R) == FLAG_R
          bytemove(@Temp, string("jmpret"), 7)
        else
          bytemove(@Temp, string("jmp"), 4)
          DestAddrVisible := FALSE

    $18:
      if (InstrFlags & FLAG_R) == FLAG_R
        bytemove(@Temp, string("and"), 4)
      else
        bytemove(@Temp, string("test"), 5)

    $19:
      bytemove(@Temp, string("andn"), 5)

    $1A:
      bytemove(@Temp, string("or"), 3)

    $1B:
      bytemove(@Temp, string("xor"), 4)

    $1C:
      bytemove(@Temp, string("muxc"), 5)

    $1D:
      bytemove(@Temp, string("muxnc"), 6)

    $1E:
      bytemove(@Temp, string("muxz"), 5)

    $1F:
      bytemove(@Temp, string("muxnz"), 6)

    $20:
      bytemove(@Temp, string("add"), 4)

    $21:
      if (InstrFlags & FLAG_R) == FLAG_R
        bytemove(@Temp, string("sub"), 4)
      else
        bytemove(@Temp, string("cmp"), 4)

    $22:
      bytemove(@Temp, string("addabs"), 7)

    $23:
      bytemove(@Temp, string("subabs"), 7)

    $24:
      bytemove(@Temp, string("sumc"), 5)

    $25:
      bytemove(@Temp, string("sumnc"), 6)

    $26:
      bytemove(@Temp, string("sumz"), 5)

    $27:
      bytemove(@Temp, string("sumnz"), 6)

    $28:
      bytemove(@Temp, string("mov"), 4)

    $29:
      bytemove(@Temp, string("neg"), 4)

    $2A:
      bytemove(@Temp, string("abs"), 4)

    $2B:
      bytemove(@Temp, string("absneg"), 7)

    $2C:
      bytemove(@Temp, string("negc"), 5)

    $2D:
      bytemove(@Temp, string("negnc"), 6)

    $2E:
      bytemove(@Temp, string("negz"), 5)

    $2F:
      bytemove(@Temp, string("negnz"), 6)

    $30:
      bytemove(@Temp, string("cmps"), 5)

    $31:
      bytemove(@Temp, string("cpmsx"), 6)

    $32:
      bytemove(@Temp, string("addx"), 5)

    $33:
      if (InstrFlags & FLAG_R) == FLAG_R
        bytemove(@Temp, string("subx"), 5)
      else
        bytemove(@Temp, string("cmpx"), 5)

    $34:
      bytemove(@Temp, string("adds"), 5)

    $35:
      bytemove(@Temp, string("subs"), 5)

    $36:
      bytemove(@Temp, string("addsx"), 6)

    $37:
      bytemove(@Temp, string("subsx"), 6)

    $38:
      bytemove(@Temp, string("cmpsub"), 7)

    $39:
      bytemove(@Temp, string("djnz"), 5)

    $3A:
      bytemove(@Temp, string("tjnz"), 5)

    $3B:
      bytemove(@Temp, string("tjz"), 4)

    $3C:
      bytemove(@Temp, string("waitpeq"), 8)

    $3D:
      bytemove(@Temp, string("waitpne"), 8)

    $3E:
      bytemove(@Temp, string("waitcnt"), 8)

    $3F:
      bytemove(@Temp, string("waitvid"), 8)

    other:
      bytemove(@Temp, string("N.A."), 5)
      DestAddrVisible := FALSE
      SrcAddrVisible := FALSE

  instruction := @Temp


PUB getDestSrcStr : destSrc | size

  size := 0
  if DestAddrVisible
    getDestinationStr
    bytemove(@Temp, @DestAddr, strsize(@DestAddr))
    size += strsize(@DestAddr)
    BYTE[@Temp][size++] := ","
    bytemove(@DestAddr, @Temp, strsize(@Temp))

  if SrcAddrVisible
    getSourceStr
    if DestAddrVisible
      bytemove(@Temp, @DestAddr, strsize(@DestAddr))

    bytemove(@Temp + size, @SrcAddr, strsize(@SrcAddr))
    size += strsize(@SrcAddr)

  BYTE[@Temp][size] := 0
  destSrc := @Temp


PUB getEffectStr : effStr | size

  size := 0
  if InstrOpcode > 3 AND (InstrOpcode == OPC_CMPS OR InstrOpcode == OPC_CMPSX)
    if (InstrFlags & FLAG_R) == FLAG_R
      bytemove(@Effect + size, @effWR, 2)
      size += 2

  if InstrOpcode > 3 AND InstrOpcode <> OPC_CALL AND InstrOpcode <> OPC_AND {
}                    AND InstrOpcode <> OPC_SUB AND InstrOpcode <> OPC_SUBX
    if (InstrFlags & FLAG_R) == 0
      bytemove(@Effect + size, @effNR, 2)
      size += 2

  if InstrFlags & FLAG_Z == FLAG_Z
    if size > 0
      BYTE[@Effect][size++] := ","

    bytemove(@Effect + size, @effZ, 2)
    size += 2

  if InstrFlags & FLAG_C == FLAG_C
    if size > 0
      BYTE[@Effect][size++] := ","

    bytemove(@Effect + size, @effC, 2)
    size += 2

  BYTE[@Effect][size] := 0

  effStr := @Effect


PUB getNextPC(pc, flags, dbgParPtr, stepInto) : nextPC | data

  case InstrCond
    0:
      InstrOpcode := 0

    1:
      if (flags & (FLAG_Z | FLAG_C)) <> 0
      InstrOpcode := 0

    2:
      if (flags & (FLAG_Z | FLAG_C)) <> FLAG_Z
        InstrOpcode := 0

    3:
      if (flags & FLAG_C) <> 0
        InstrOpcode := 0

    4:
      if (flags & (FLAG_Z | FLAG_C)) <> FLAG_C
        InstrOpcode := 0

    5:
      if (flags & FLAG_Z) <> 0
        InstrOpcode := 0

    6:
      if (flags & FLAG_Z) == (flags & FLAG_C) << 1
        InstrOpcode := 0

    7:
      if (flags & FLAG_Z) <> 0 OR (flags & FLAG_C) <> 0
        InstrOpcode := 0

    8:
      if (flags & (FLAG_Z | FLAG_C)) <> (FLAG_Z | FLAG_C)
        InstrOpcode := 0

    9:
      if (flags & FLAG_Z) <> (flags & FLAG_C) << 1
        InstrOpcode := 0

    10:
      if (flags & FLAG_Z) <> FLAG_Z
        InstrOpcode := 0

    11:
      if (flags & FLAG_Z) <> FLAG_Z OR (flags & FLAG_C) <> 0
        InstrOpcode := 0

    12:
      if (flags & FLAG_C) <> FLAG_C
        InstrOpcode := 0

    13:
      if (flags & FLAG_Z) <> 0 OR (flags & FLAG_C) <> FLAG_C
        InstrOpcode := 0

    14:
      if (flags & FLAG_Z) <> FLAG_Z OR (flags & FLAG_C) <> FLAG_C
        InstrOpcode := 0

  case InstrOpcode
    $17: 'call, ret, jmpret, jmp
      if (InstrFlags & FLAG_I) == FLAG_I
        if (InstrFlags & FLAG_R) == FLAG_R              'call
          if stepInto
            nextPC := InstrSrc
          else
            nextPC := pc + 1
        else                                            'ret
          nextPC := InstrSrc
      else
        if (InstrFlags & FLAG_R) == FLAG_R              'jmpret
          if stepInto
            data := getCogData(dbgParPtr, InstrSrc)
            nextPC := data
          else
            nextPC := pc + 1
        else                                            'jmp
          nextPC := InstrSrc

    $39: 'djnz
      data := getCogData(dbgParPtr, InstrDest)
      if (data - 1) <> 0
        nextPC := InstrSrc
      else
        nextPC := pc + 1

    $3A: 'tjnz
      data := getCogData(dbgParPtr, InstrDest)
      if data <> 0
        nextPC := InstrSrc
      else
        nextPC := pc + 1

    $3B: 'tjz
      data := getCogData(dbgParPtr, InstrDest)
      if data == 0
        nextPC := InstrSrc
      else
        nextPC := pc + 1

    other:
      nextPC := pc + 1


PUB getPrintableStr(value, type) : printable | byteValue, i

  repeat i from 0 to type
    byteValue := (value & $FF)
    value >>= 8
    case byteValue
      $20..$7F:
        BYTE[@Temp][i] := byteValue

      other:
        BYTE[@Temp][i] := "."

  BYTE[@Temp][type + 1] := 0
  printable := @Temp


PRI getDestinationStr : destination | size

  size := 0
  if DestAddrVisible
    hex(InstrDest, 3)
    BYTE[@DestAddr][size++] := "$"
    bytemove(@DestAddr + size, @Temp, strsize(@Temp))
    size += strsize(@Temp)

  BYTE[@DestAddr][size] := 0
  destination := @DestAddr


PRI getSourceStr : source | size

  size := 0
  if SrcAddrVisible
    hex(InstrSrc, 3)
    if (InstrFlags & FLAG_I) == FLAG_I
      BYTE[@SrcAddr][size++] := "#"

    BYTE[@SrcAddr][size++] := "$"
    bytemove(@SrcAddr + size, @Temp, strsize(@Temp))
    size += strsize(@Temp)

  BYTE[@SrcAddr][size] := 0
  source := @SrcAddr


PRI hex(value, digits) | i

  value <<= (8 - digits) << 2
  repeat i from 1 to digits
    BYTE[@Temp][i - 1] := lookupz((value <-= 4) & $F : "0".."9", "A".."F")

  BYTE[@Temp][digits] := 0


PRI getCogData(dbgParPtr, addr) : data

  LONG[dbgParPtr][1] := addr
  LONG[dbgParPtr] := CMD_RDCOGDATA
  repeat until LONG[dbgParPtr] == 0

  data := LONG[dbgParPtr][2]


DAT

effNR         byte  "nr",0
effWR         byte  "wr",0
effC          byte  "wc",0
effZ          byte  "wz",0

conditionTable
never_always  byte  0         'never or always
ifNZaNC       byte  "NZ&NC",0
ifZaNC        byte  "Z&NC",0
ifNC          byte  "NC",0
ifNZaC        byte  "NZ&C",0
ifNZ          byte  "NZ",0
ifZneC        byte  "Z<>C",0
ifNZoNC       byte  "NZ/NC",0
ifZaC         byte  "Z&C",0
ifZeqC        byte  "Z=C",0
ifZ           byte  "Z",0
ifZoNC        byte  "Z/NC",0
ifC           byte  "C",0
ifNZoC        byte  "NZ/C",0
ifZoC         byte  "Z/C",0
'